@@ -39,7 +39,7 @@
 
 /**
  * gnutls_random_art:
- * @type: The type of the random art
+ * @type: The type of the random art (for now only %GNUTLS_RANDOM_ART_OPENSSH is supported)
  * @key_type: The type of the key (RSA, DSA etc.)
  * @key_size: The size of the key in bits
  * @fpr: The fingerprint of the key
@@ -47,7 +47,7 @@
  * @art: The returned random art
  *
  * This function will convert a given fingerprint to an "artistic"
- * image. The returned image is allocated using gnutls_malloc()
+ * image. The returned image is allocated using gnutls_malloc().
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise
  *   an error code is returned.
@@ -77,7 +77,7 @@ int gnutls_random_art(gnutls_random_art_t type,
 #if defined(ENABLE_DHE) || defined(ENABLE_ANON)
 /**
  * gnutls_dh_set_prime_bits:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  * @bits: is the number of bits
  *
  * This function sets the number of bits, for use in a Diffie-Hellman
@@ -103,7 +103,8 @@ int gnutls_random_art(gnutls_random_art_t type,
  **/
 void gnutls_dh_set_prime_bits(gnutls_session_t session, unsigned int bits)
 {
-	if (bits <= 512 && bits != 0)
+	if (bits <= gnutls_sec_param_to_pk_bits(GNUTLS_PK_DH, GNUTLS_SEC_PARAM_VERY_WEAK)
+		&& bits != 0)
 		_gnutls_audit_log(session,
 				  "Note that the security level of the Diffie-Hellman key exchange has been lowered to %u bits and this may allow decryption of the session data\n",
 				  bits);
@@ -138,19 +139,19 @@ gnutls_dh_get_group(gnutls_session_t session,
 
 	switch (gnutls_auth_get_type(session)) {
 	case GNUTLS_CRD_ANON:
-		anon_info = _gnutls_get_auth_info(session);
+		anon_info = _gnutls_get_auth_info(session, GNUTLS_CRD_ANON);
 		if (anon_info == NULL)
 			return GNUTLS_E_INTERNAL_ERROR;
 		dh = &anon_info->dh;
 		break;
 	case GNUTLS_CRD_PSK:
-		psk_info = _gnutls_get_auth_info(session);
+		psk_info = _gnutls_get_auth_info(session, GNUTLS_CRD_PSK);
 		if (psk_info == NULL)
 			return GNUTLS_E_INTERNAL_ERROR;
 		dh = &psk_info->dh;
 		break;
 	case GNUTLS_CRD_CERTIFICATE:
-		cert_info = _gnutls_get_auth_info(session);
+		cert_info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 		if (cert_info == NULL)
 			return GNUTLS_E_INTERNAL_ERROR;
 		dh = &cert_info->dh;
@@ -202,7 +203,7 @@ gnutls_dh_get_pubkey(gnutls_session_t session, gnutls_datum_t * raw_key)
 	switch (gnutls_auth_get_type(session)) {
 	case GNUTLS_CRD_ANON:
 		{
-			anon_info = _gnutls_get_auth_info(session);
+			anon_info = _gnutls_get_auth_info(session, GNUTLS_CRD_ANON);
 			if (anon_info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			dh = &anon_info->dh;
@@ -210,7 +211,7 @@ gnutls_dh_get_pubkey(gnutls_session_t session, gnutls_datum_t * raw_key)
 		}
 	case GNUTLS_CRD_PSK:
 		{
-			psk_info = _gnutls_get_auth_info(session);
+			psk_info = _gnutls_get_auth_info(session, GNUTLS_CRD_PSK);
 			if (psk_info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			dh = &psk_info->dh;
@@ -219,7 +220,7 @@ gnutls_dh_get_pubkey(gnutls_session_t session, gnutls_datum_t * raw_key)
 	case GNUTLS_CRD_CERTIFICATE:
 		{
 
-			cert_info = _gnutls_get_auth_info(session);
+			cert_info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 			if (cert_info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			dh = &cert_info->dh;
@@ -252,7 +253,7 @@ int gnutls_dh_get_secret_bits(gnutls_session_t session)
 		{
 			anon_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_ANON);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			return info->dh.secret_bits;
@@ -261,7 +262,7 @@ int gnutls_dh_get_secret_bits(gnutls_session_t session)
 		{
 			psk_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_PSK);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			return info->dh.secret_bits;
@@ -270,7 +271,7 @@ int gnutls_dh_get_secret_bits(gnutls_session_t session)
 		{
 			cert_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 
@@ -288,7 +289,7 @@ static int mpi_buf2bits(gnutls_datum_t * mpi_buf)
 	bigint_t mpi;
 	int rc;
 
-	rc = _gnutls_mpi_scan_nz(&mpi, mpi_buf->data, mpi_buf->size);
+	rc = _gnutls_mpi_init_scan_nz(&mpi, mpi_buf->data, mpi_buf->size);
 	if (rc) {
 		gnutls_assert();
 		return rc;
@@ -323,7 +324,7 @@ int gnutls_dh_get_prime_bits(gnutls_session_t session)
 		{
 			anon_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_ANON);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			dh = &info->dh;
@@ -333,7 +334,7 @@ int gnutls_dh_get_prime_bits(gnutls_session_t session)
 		{
 			psk_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_PSK);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 			dh = &info->dh;
@@ -343,7 +344,7 @@ int gnutls_dh_get_prime_bits(gnutls_session_t session)
 		{
 			cert_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 
@@ -378,7 +379,7 @@ int gnutls_dh_get_peers_public_bits(gnutls_session_t session)
 		{
 			anon_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_ANON);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 
@@ -389,7 +390,7 @@ int gnutls_dh_get_peers_public_bits(gnutls_session_t session)
 		{
 			psk_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_PSK);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 
@@ -400,7 +401,7 @@ int gnutls_dh_get_peers_public_bits(gnutls_session_t session)
 		{
 			cert_auth_info_t info;
 
-			info = _gnutls_get_auth_info(session);
+			info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 			if (info == NULL)
 				return GNUTLS_E_INTERNAL_ERROR;
 
@@ -417,8 +418,8 @@ int gnutls_dh_get_peers_public_bits(gnutls_session_t session)
 
 /**
  * gnutls_certificate_set_dh_params:
- * @res: is a gnutls_certificate_credentials_t structure
- * @dh_params: is a structure that holds Diffie-Hellman parameters.
+ * @res: is a gnutls_certificate_credentials_t type
+ * @dh_params: the Diffie-Hellman parameters.
  *
  * This function will set the Diffie-Hellman parameters for a
  * certificate server to use. These parameters will be used in
@@ -457,7 +458,7 @@ const gnutls_datum_t *gnutls_certificate_get_ours(gnutls_session_t session)
 	CHECK_AUTH(GNUTLS_CRD_CERTIFICATE, NULL);
 
 	cred = (gnutls_certificate_credentials_t)
-	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE, NULL);
+	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
 	if (cred == NULL || cred->certs == NULL) {
 		gnutls_assert();
 		return NULL;
@@ -472,7 +473,7 @@ const gnutls_datum_t *gnutls_certificate_get_ours(gnutls_session_t session)
 /**
  * gnutls_certificate_get_peers:
  * @session: is a gnutls session
- * @list_size: is the length of the certificate list
+ * @list_size: is the length of the certificate list (may be %NULL)
  *
  * Get the peer's raw certificate (chain) as sent by the peer.  These
  * certificates are in raw format (DER encoded for X.509).  In case of
@@ -483,7 +484,7 @@ const gnutls_datum_t *gnutls_certificate_get_ours(gnutls_session_t session)
  * In case of OpenPGP keys a single key will be returned in raw
  * format.
  *
- * Returns: a pointer to a #gnutls_datum_t containing our
+ * Returns: a pointer to a #gnutls_datum_t containing the peer's
  *   certificates, or %NULL in case of an error or if no certificate
  *   was used.
  **/
@@ -495,11 +496,12 @@ const gnutls_datum_t *gnutls_certificate_get_peers(gnutls_session_t
 
 	CHECK_AUTH(GNUTLS_CRD_CERTIFICATE, NULL);
 
-	info = _gnutls_get_auth_info(session);
+	info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 	if (info == NULL)
 		return NULL;
 
-	*list_size = info->ncerts;
+	if (list_size)
+		*list_size = info->ncerts;
 	return info->raw_certificate_list;
 }
 
@@ -524,7 +526,7 @@ int gnutls_certificate_get_peers_subkey_id(gnutls_session_t session,
 
 	CHECK_AUTH(GNUTLS_CRD_CERTIFICATE, GNUTLS_E_INVALID_REQUEST);
 
-	info = _gnutls_get_auth_info(session);
+	info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
 	if (info == NULL)
 		return GNUTLS_E_INVALID_REQUEST;
 
@@ -542,8 +544,7 @@ int gnutls_certificate_get_peers_subkey_id(gnutls_session_t session,
  * Get whether client certificate is requested or not.
  *
  * Returns: 0 if the peer (server) did not request client
- *   authentication or 1 otherwise, or a negative error code in case of
- *   error.
+ *   authentication or 1 otherwise.
  **/
 int gnutls_certificate_client_get_request_status(gnutls_session_t session)
 {
@@ -576,7 +577,7 @@ gnutls_fingerprint(gnutls_digest_algorithm_t algo,
 		   size_t * result_size)
 {
 	int ret;
-	int hash_len = _gnutls_hash_get_algo_len(mac_to_entry(algo));
+	int hash_len = _gnutls_hash_get_algo_len(hash_to_entry(algo));
 
 	if (hash_len < 0 || (unsigned) hash_len > *result_size
 	    || result == NULL) {
@@ -598,7 +599,7 @@ gnutls_fingerprint(gnutls_digest_algorithm_t algo,
 
 /**
  * gnutls_certificate_set_params_function:
- * @res: is a gnutls_certificate_credentials_t structure
+ * @res: is a gnutls_certificate_credentials_t type
  * @func: is the function to be called
  *
  * This function will set a callback in order for the server to get
@@ -614,7 +615,7 @@ gnutls_certificate_set_params_function(gnutls_certificate_credentials_t
 
 /**
  * gnutls_certificate_set_verify_flags:
- * @res: is a gnutls_certificate_credentials_t structure
+ * @res: is a gnutls_certificate_credentials_t type
  * @flags: are the flags
  *
  * This function will set the flags to be used for verification 
@@ -629,9 +630,26 @@ gnutls_certificate_set_verify_flags(gnutls_certificate_credentials_t
 	res->verify_flags = flags;
 }
 
+/**
+ * gnutls_certificate_get_verify_flags:
+ * @res: is a gnutls_certificate_credentials_t type
+ *
+ * Returns the verification flags set with
+ * gnutls_certificate_set_verify_flags().
+ *
+ * Returns: The certificate verification flags used by @res.
+ *
+ * Since: 3.4.0
+ */
+unsigned int
+gnutls_certificate_get_verify_flags(gnutls_certificate_credentials_t res)
+{
+	return res->verify_flags;
+}
+
 /**
  * gnutls_certificate_set_verify_limits:
- * @res: is a gnutls_certificate_credentials structure
+ * @res: is a gnutls_certificate_credentials type
  * @max_bits: is the number of bits of an acceptable certificate (default 8200)
  * @max_depth: is maximum depth of the verification of a certificate chain (default 5)
  *
@@ -652,7 +670,7 @@ gnutls_certificate_set_verify_limits(gnutls_certificate_credentials_t res,
 #ifdef ENABLE_PSK
 /**
  * gnutls_psk_set_params_function:
- * @res: is a gnutls_psk_server_credentials_t structure
+ * @res: is a gnutls_psk_server_credentials_t type
  * @func: is the function to be called
  *
  * This function will set a callback in order for the server to get
@@ -670,7 +688,7 @@ gnutls_psk_set_params_function(gnutls_psk_server_credentials_t res,
 #ifdef ENABLE_ANON
 /**
  * gnutls_anon_set_params_function:
- * @res: is a gnutls_anon_server_credentials_t structure
+ * @res: is a gnutls_anon_server_credentials_t type
  * @func: is the function to be called
  *
  * This function will set a callback in order for the server to get
@@ -720,80 +738,43 @@ int gnutls_load_file(const char *filename, gnutls_datum_t * data)
 	return 0;
 }
 
+#ifdef ENABLE_OCSP
 /**
  * gnutls_ocsp_status_request_is_checked:
  * @session: is a gnutls session
- * @flags: should be zero
+ * @flags: should be zero or %GNUTLS_OCSP_SR_IS_AVAIL
  *
  * Check whether an OCSP status response was included in the handshake
  * and whether it was checked and valid (not too old or superseded). 
  * This is a helper function when needing to decide whether to perform an
- * OCSP validity check on the peer's certificate. Must be called after
- * gnutls_certificate_verify_peers3() is called.
+ * OCSP validity check on the peer's certificate. Should be called after
+ * any of gnutls_certificate_verify_peers*() are called.
  *
- * Returns: non zero it was valid, or a zero if it wasn't sent,
+ * If the flag %GNUTLS_OCSP_SR_IS_AVAIL is specified, the return
+ * value of the function indicates whether an OCSP status response have
+ * been received (even if invalid).
+ *
+ * Returns: non zero if the response was valid, or a zero if it wasn't sent,
  * or sent and was invalid.
  **/
 int
 gnutls_ocsp_status_request_is_checked(gnutls_session_t session,
 				      unsigned int flags)
 {
+	int ret;
+	gnutls_datum_t data;
+
+	if (flags & GNUTLS_OCSP_SR_IS_AVAIL) {
+		ret = gnutls_ocsp_status_request_get(session, &data);
+		if (ret < 0)
+			return gnutls_assert_val(0);
+
+		if (data.data == NULL)
+			return gnutls_assert_val(0);
+		return 1;
+	}
 	return session->internals.ocsp_check_ok;
 }
-
-#ifdef ENABLE_RSA_EXPORT
-
-/**
- * gnutls_rsa_export_get_pubkey:
- * @session: is a gnutls session
- * @exponent: will hold the exponent.
- * @modulus: will hold the modulus.
- *
- * This function will return the peer's public key exponent and
- * modulus used in the last RSA-EXPORT authentication.  The output
- * parameters must be freed with gnutls_free().
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise
- *   an error code is returned.
- **/
-int
-gnutls_rsa_export_get_pubkey(gnutls_session_t session,
-			     gnutls_datum_t * exponent,
-			     gnutls_datum_t * modulus)
-{
-	return gnutls_assert_val(GNUTLS_E_UNIMPLEMENTED_FEATURE);
-}
-
-/**
- * gnutls_rsa_export_get_modulus_bits:
- * @session: is a gnutls session
- *
- * Get the export RSA parameter's modulus size.
- *
- * Returns: The bits used in the last RSA-EXPORT key exchange with the
- *   peer, or a negative error code in case of error.
- **/
-int gnutls_rsa_export_get_modulus_bits(gnutls_session_t session)
-{
-	return gnutls_assert_val(GNUTLS_E_UNIMPLEMENTED_FEATURE);
-}
-
-/**
- * gnutls_certificate_set_rsa_export_params:
- * @res: is a gnutls_certificate_credentials_t structure
- * @rsa_params: is a structure that holds temporary RSA parameters.
- *
- * This function will set the temporary RSA parameters for a
- * certificate server to use.  These parameters will be used in
- * RSA-EXPORT cipher suites.
- **/
-void
-gnutls_certificate_set_rsa_export_params(gnutls_certificate_credentials_t
-					 res,
-					 gnutls_rsa_params_t rsa_params)
-{
-	return;
-}
 #endif
 
 #define DESC_SIZE 64
@@ -805,6 +786,9 @@ gnutls_certificate_set_rsa_export_params(gnutls_certificate_credentials_t
  * This function returns a string describing the current session.
  * The string is null terminated and allocated using gnutls_malloc().
  *
+ * If initial negotiation is not complete when this function is called,
+ * %NULL will be returned.
+ *
  * Returns: a description of the protocols and algorithms in the current session.
  *
  * Since: 3.1.10
@@ -812,13 +796,18 @@ gnutls_certificate_set_rsa_export_params(gnutls_certificate_credentials_t
 char *gnutls_session_get_desc(gnutls_session_t session)
 {
 	gnutls_kx_algorithm_t kx;
+	const char *kx_str;
 	unsigned type;
 	char kx_name[32];
 	char proto_name[32];
 	const char *curve_name = NULL;
 	unsigned dh_bits = 0;
+	unsigned mac_id;
 	char *desc;
 
+	if (session->internals.initial_negotiation_completed == 0)
+		return NULL;
+
 	kx = session->security_parameters.kx_algorithm;
 
 	if (kx == GNUTLS_KX_ANON_ECDH || kx == GNUTLS_KX_ECDHE_PSK ||
@@ -826,24 +815,31 @@ char *gnutls_session_get_desc(gnutls_session_t session)
 		curve_name =
 		    gnutls_ecc_curve_get_name(gnutls_ecc_curve_get
 					      (session));
+#if defined(ENABLE_DHE) || defined(ENABLE_ANON)
 	} else if (kx == GNUTLS_KX_ANON_DH || kx == GNUTLS_KX_DHE_PSK
 		   || kx == GNUTLS_KX_DHE_RSA || kx == GNUTLS_KX_DHE_DSS) {
 		dh_bits = gnutls_dh_get_prime_bits(session);
+#endif
 	}
 
-	if (curve_name != NULL)
-		snprintf(kx_name, sizeof(kx_name), "%s-%s",
-			 gnutls_kx_get_name(kx), curve_name);
-	else if (dh_bits != 0)
-		snprintf(kx_name, sizeof(kx_name), "%s-%u",
-			 gnutls_kx_get_name(kx), dh_bits);
-	else
-		snprintf(kx_name, sizeof(kx_name), "%s",
-			 gnutls_kx_get_name(kx));
+	kx_str = gnutls_kx_get_name(kx);
+	if (kx_str) {
+		if (curve_name != NULL)
+			snprintf(kx_name, sizeof(kx_name), "%s-%s",
+				 kx_str, curve_name);
+		else if (dh_bits != 0)
+			snprintf(kx_name, sizeof(kx_name), "%s-%u",
+				 kx_str, dh_bits);
+		else
+			snprintf(kx_name, sizeof(kx_name), "%s",
+				 kx_str);
+	} else {
+		strcpy(kx_name, "NULL");
+	}
 
 	type = gnutls_certificate_type_get(session);
 	if (type == GNUTLS_CRT_X509)
-		snprintf(proto_name, sizeof(proto_name), "%s-PKIX",
+		snprintf(proto_name, sizeof(proto_name), "%s",
 			 gnutls_protocol_get_name(get_num_version
 						  (session)));
 	else
@@ -857,19 +853,28 @@ char *gnutls_session_get_desc(gnutls_session_t session)
 	if (desc == NULL)
 		return NULL;
 
-	snprintf(desc, DESC_SIZE,
-		 "(%s)-(%s)-(%s)-(%s)",
-		 proto_name,
-		 kx_name,
-		 gnutls_cipher_get_name(gnutls_cipher_get(session)),
-		 gnutls_mac_get_name(gnutls_mac_get(session)));
+	mac_id = gnutls_mac_get(session);
+	if (mac_id == GNUTLS_MAC_AEAD) { /* no need to print */
+		snprintf(desc, DESC_SIZE,
+			 "(%s)-(%s)-(%s)",
+			 proto_name,
+			 kx_name,
+			 gnutls_cipher_get_name(gnutls_cipher_get(session)));
+	} else {
+		snprintf(desc, DESC_SIZE,
+			 "(%s)-(%s)-(%s)-(%s)",
+			 proto_name,
+			 kx_name,
+			 gnutls_cipher_get_name(gnutls_cipher_get(session)),
+			 gnutls_mac_get_name(mac_id));
+	}
 
 	return desc;
 }
 
 /**
  * gnutls_session_set_id:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  * @sid: the session identifier
  *
  * This function sets the session ID to be used in a client hello.
@@ -885,7 +890,7 @@ int
 gnutls_session_set_id(gnutls_session_t session, const gnutls_datum_t * sid)
 {
 	if (session->security_parameters.entity == GNUTLS_SERVER ||
-	    sid->size > TLS_MAX_SESSION_ID_SIZE)
+	    sid->size > GNUTLS_MAX_SESSION_ID_SIZE)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
 	memset(&session->internals.resumed_security_parameters, 0,
