@@ -32,12 +32,13 @@
 #include <gnutls_pk.h>
 #include <gnutls_mpi.h>
 #include <gnutls_ecc.h>
+#include <pin.h>
 
 /**
  * gnutls_x509_privkey_init:
- * @key: The structure to be initialized
+ * @key: A pointer to the type to be initialized
  *
- * This function will initialize an private key structure.
+ * This function will initialize a private key type.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
@@ -57,9 +58,17 @@ int gnutls_x509_privkey_init(gnutls_x509_privkey_t * key)
 	return GNUTLS_E_MEMORY_ERROR;
 }
 
+void _gnutls_x509_privkey_reinit(gnutls_x509_privkey_t key)
+{
+	gnutls_pk_params_clear(&key->params);
+	gnutls_pk_params_release(&key->params);
+	asn1_delete_structure2(&key->key, ASN1_DELETE_FLAG_ZEROIZE);
+	key->key = ASN1_TYPE_EMPTY;
+}
+
 /**
  * gnutls_x509_privkey_deinit:
- * @key: The structure to be deinitialized
+ * @key: The key to be deinitialized
  *
  * This function will deinitialize a private key structure.
  **/
@@ -68,9 +77,7 @@ void gnutls_x509_privkey_deinit(gnutls_x509_privkey_t key)
 	if (!key)
 		return;
 
-	gnutls_pk_params_clear(&key->params);
-	gnutls_pk_params_release(&key->params);
-	asn1_delete_structure2(&key->key, ASN1_DELETE_FLAG_ZEROIZE);
+	_gnutls_x509_privkey_reinit(key);
 	gnutls_free(key);
 }
 
@@ -141,7 +148,7 @@ _gnutls_privkey_decode_pkcs1_rsa_key(const gnutls_datum_t * raw_key,
 	}
 
 	result =
-	    asn1_der_decoding(&pkey_asn, raw_key->data, raw_key->size,
+	    _asn1_strict_der_decode(&pkey_asn, raw_key->data, raw_key->size,
 			      NULL);
 	if (result != ASN1_SUCCESS) {
 		gnutls_assert();
@@ -256,7 +263,7 @@ _gnutls_privkey_decode_ecc_key(ASN1_TYPE* pkey_asn, const gnutls_datum_t * raw_k
 	}
 
 	ret =
-	    asn1_der_decoding(pkey_asn, raw_key->data, raw_key->size,
+	    _asn1_strict_der_decode(pkey_asn, raw_key->data, raw_key->size,
 			      NULL);
 	if (ret != ASN1_SUCCESS) {
 		gnutls_assert();
@@ -363,7 +370,7 @@ decode_dsa_key(const gnutls_datum_t * raw_key, gnutls_x509_privkey_t pkey)
 	pkey->params.algo = GNUTLS_PK_DSA;
 
 	result =
-	    asn1_der_decoding(&dsa_asn, raw_key->data, raw_key->size,
+	    _asn1_strict_der_decode(&dsa_asn, raw_key->data, raw_key->size,
 			      NULL);
 	if (result != ASN1_SUCCESS) {
 		gnutls_assert();
@@ -427,7 +434,7 @@ decode_dsa_key(const gnutls_datum_t * raw_key, gnutls_x509_privkey_t pkey)
 
 /**
  * gnutls_x509_privkey_import:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @data: The DER or PEM encoded certificate.
  * @format: One of DER or PEM
  *
@@ -501,6 +508,11 @@ gnutls_x509_privkey_import(gnutls_x509_privkey_t key,
 		need_free = 1;
 	}
 
+	if (key->expanded) {
+		_gnutls_x509_privkey_reinit(key);
+	}
+	key->expanded = 1;
+
 	if (key->pk_algorithm == GNUTLS_PK_RSA) {
 		key->key =
 		    _gnutls_privkey_decode_pkcs1_rsa_key(&_data, key);
@@ -614,7 +626,7 @@ static int import_pkcs12_privkey(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_import2:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @data: The DER or PEM encoded key.
  * @format: One of DER or PEM
  * @password: A password (optional)
@@ -628,7 +640,8 @@ static int import_pkcs12_privkey(gnutls_x509_privkey_t key,
  * and the openssl format.
  *
  * If the provided key is encrypted but no password was given, then
- * %GNUTLS_E_DECRYPTION_FAILED is returned.
+ * %GNUTLS_E_DECRYPTION_FAILED is returned. Since GnuTLS 3.4.0 this
+ * function will utilize the PIN callbacks if any.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
@@ -640,6 +653,7 @@ gnutls_x509_privkey_import2(gnutls_x509_privkey_t key,
 			    const char *password, unsigned int flags)
 {
 	int ret = 0;
+	char pin[GNUTLS_PKCS11_MAX_PIN_LEN];
 
 	if (password == NULL && !(flags & GNUTLS_PKCS_NULL_PASSWORD)) {
 		ret = gnutls_x509_privkey_import(key, data, format);
@@ -650,9 +664,24 @@ gnutls_x509_privkey_import2(gnutls_x509_privkey_t key,
 
 	if ((password != NULL || (flags & GNUTLS_PKCS_NULL_PASSWORD))
 	    || ret < 0) {
+
 		ret =
 		    gnutls_x509_privkey_import_pkcs8(key, data, format,
 						     password, flags);
+
+		if (ret == GNUTLS_E_DECRYPTION_FAILED &&
+		    password == NULL && (!(flags & GNUTLS_PKCS_PLAIN))) {
+		    /* use the callback if any */
+			ret = _gnutls_retrieve_pin(&key->pin, "key:", "", 0, pin, sizeof(pin));
+			if (ret == 0) {
+		    		password = pin;
+			}
+
+			ret =
+			    gnutls_x509_privkey_import_pkcs8(key, data, format,
+						     password, flags);
+		}
+
 		if (ret < 0) {
 			if (ret == GNUTLS_E_DECRYPTION_FAILED)
 				goto cleanup;
@@ -687,7 +716,7 @@ gnutls_x509_privkey_import2(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_import_rsa_raw:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @m: holds the modulus
  * @e: holds the public exponent
  * @d: holds the private exponent
@@ -717,7 +746,7 @@ gnutls_x509_privkey_import_rsa_raw(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_import_rsa_raw2:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @m: holds the modulus
  * @e: holds the public exponent
  * @d: holds the private exponent
@@ -853,7 +882,7 @@ gnutls_x509_privkey_import_rsa_raw2(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_import_dsa_raw:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @p: holds the p
  * @q: holds the q
  * @g: holds the g
@@ -940,7 +969,7 @@ gnutls_x509_privkey_import_dsa_raw(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_import_ecc_raw:
- * @key: The structure to store the parsed key
+ * @key: The data to store the parsed key
  * @curve: holds the curve
  * @x: holds the x
  * @y: holds the y
@@ -1009,7 +1038,7 @@ gnutls_x509_privkey_import_ecc_raw(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_get_pk_algorithm:
- * @key: should contain a #gnutls_x509_privkey_t structure
+ * @key: should contain a #gnutls_x509_privkey_t type
  *
  * This function will return the public key algorithm of a private
  * key.
@@ -1029,7 +1058,7 @@ int gnutls_x509_privkey_get_pk_algorithm(gnutls_x509_privkey_t key)
 
 /**
  * gnutls_x509_privkey_get_pk_algorithm2:
- * @key: should contain a #gnutls_x509_privkey_t structure
+ * @key: should contain a #gnutls_x509_privkey_t type
  * @bits: The number of bits in the public key algorithm
  *
  * This function will return the public key algorithm of a private
@@ -1150,7 +1179,7 @@ gnutls_x509_privkey_export2(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_sec_param:
- * @key: a key structure
+ * @key: a key
  *
  * This function will return the security parameter appropriate with
  * this private key.
@@ -1173,7 +1202,7 @@ gnutls_sec_param_t gnutls_x509_privkey_sec_param(gnutls_x509_privkey_t key)
 
 /**
  * gnutls_x509_privkey_export_ecc_raw:
- * @key: a structure that holds the rsa parameters
+ * @key: a key
  * @curve: will hold the curve
  * @x: will hold the x coordinate
  * @y: will hold the y coordinate
@@ -1204,7 +1233,7 @@ int gnutls_x509_privkey_export_ecc_raw(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_export_rsa_raw:
- * @key: a structure that holds the rsa parameters
+ * @key: a key
  * @m: will hold the modulus
  * @e: will hold the public exponent
  * @d: will hold the private exponent
@@ -1230,7 +1259,7 @@ gnutls_x509_privkey_export_rsa_raw(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_export_rsa_raw2:
- * @key: a structure that holds the rsa parameters
+ * @key: a key
  * @m: will hold the modulus
  * @e: will hold the public exponent
  * @d: will hold the private exponent
@@ -1262,7 +1291,7 @@ gnutls_x509_privkey_export_rsa_raw2(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_export_dsa_raw:
- * @key: a structure that holds the DSA parameters
+ * @key: a key
  * @p: will hold the p
  * @q: will hold the q
  * @g: will hold the g
@@ -1316,12 +1345,19 @@ char* gen_data = NULL;
 			goto cleanup;
 		}
 
+		if (ddata.size == sig.size && memcmp(ddata.data, sig.data, sig.size) == 0) {
+			ret = gnutls_assert_val(GNUTLS_E_PK_GENERATION_ERROR);
+			gnutls_assert();
+			goto cleanup;
+		}
+
 		ret = _gnutls_pk_decrypt(algo, &tmp, &sig, params);
 		if (ret < 0) {
 			ret = gnutls_assert_val(GNUTLS_E_PK_GENERATION_ERROR);
 			gnutls_assert();
 			goto cleanup;
 		}
+
 		if (tmp.size != ddata.size || memcmp(tmp.data, ddata.data, tmp.size) != 0) {
 			ret = gnutls_assert_val(GNUTLS_E_PK_GENERATION_ERROR);
 			gnutls_assert();
@@ -1357,6 +1393,9 @@ char* gen_data = NULL;
 
 	ret = 0;
 cleanup:
+	if (ret == GNUTLS_E_PK_GENERATION_ERROR) {
+		_gnutls_switch_lib_state(LIB_STATE_ERROR);
+	}
 	gnutls_free(gen_data);
 	gnutls_free(sig.data);
 	gnutls_free(tmp.data);
@@ -1366,7 +1405,7 @@ cleanup:
 
 /**
  * gnutls_x509_privkey_generate:
- * @key: should contain a #gnutls_x509_privkey_t structure
+ * @key: a key
  * @algo: is one of the algorithms in #gnutls_pk_algorithm_t.
  * @bits: the size of the modulus
  * @flags: unused for now.  Must be 0.
@@ -1416,7 +1455,7 @@ gnutls_x509_privkey_generate(gnutls_x509_privkey_t key,
 	ret = _gnutls_pk_generate_keys(algo, bits, &key->params);
 	if (ret < 0) {
 		gnutls_assert();
-		return ret;
+		goto cleanup;
 	}
 
 #ifndef ENABLE_FIPS140
@@ -1426,7 +1465,7 @@ gnutls_x509_privkey_generate(gnutls_x509_privkey_t key,
 #endif
 	if (ret < 0) {
 		gnutls_assert();
-		return ret;
+		goto cleanup;
 	}
 
 	ret = _gnutls_asn1_encode_privkey(algo, &key->key, &key->params);
@@ -1448,7 +1487,7 @@ gnutls_x509_privkey_generate(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_verify_params:
- * @key: should contain a #gnutls_x509_privkey_t structure
+ * @key: a key
  *
  * This function will verify the private key parameters.
  *
@@ -1470,7 +1509,7 @@ int gnutls_x509_privkey_verify_params(gnutls_x509_privkey_t key)
 
 /**
  * gnutls_x509_privkey_get_key_id:
- * @key: Holds the key
+ * @key: a key
  * @flags: should be 0 for now
  * @output_data: will contain the key ID
  * @output_data_size: holds the size of output_data (and will be
@@ -1576,7 +1615,7 @@ _gnutls_x509_privkey_sign_hash2(gnutls_x509_privkey_t signer,
 
 /**
  * gnutls_x509_privkey_sign_hash:
- * @key: Holds the key
+ * @key: a key
  * @hash: holds the data to be signed
  * @signature: will contain newly allocated signature
  *
@@ -1616,7 +1655,7 @@ gnutls_x509_privkey_sign_hash(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_sign_data:
- * @key: Holds the key
+ * @key: a key
  * @digest: should be MD5 or SHA1
  * @flags: should be 0 for now
  * @data: holds the data to be signed
@@ -1692,7 +1731,7 @@ gnutls_x509_privkey_sign_data(gnutls_x509_privkey_t key,
 
 /**
  * gnutls_x509_privkey_fix:
- * @key: Holds the key
+ * @key: a key
  *
  * This function will recalculate the secondary parameters in a key.
  * In RSA keys, this can be the coefficient and exponent1,2.
@@ -1721,3 +1760,26 @@ int gnutls_x509_privkey_fix(gnutls_x509_privkey_t key)
 
 	return 0;
 }
+
+/**
+ * gnutls_x509_privkey_set_pin_function:
+ * @privkey: The certificate structure
+ * @fn: the callback
+ * @userdata: data associated with the callback
+ *
+ * This function will set a callback function to be used when
+ * it is required to access a protected object. This function overrides 
+ * the global function set using gnutls_pkcs11_set_pin_function().
+ *
+ * Note that this callback is used when decrypting a key.
+ *
+ * Since: 3.4.0
+ *
+ **/
+void gnutls_x509_privkey_set_pin_function(gnutls_x509_privkey_t privkey,
+				      gnutls_pin_callback_t fn,
+				      void *userdata)
+{
+	privkey->pin.cb = fn;
+	privkey->pin.data = userdata;
+}
